'Attribute VB_Name = "KeymanLink"
'###########################################################################################
'
'  VBA Interface for enabling and disabling Keyman 5 (or later) keyboards
'  ======================================================================

'  Usage:
'    To enable any keyboard, simply call:
'          SelectKeymanKeyboard "<name>"
'    where <name> is the name of the keyboard shown in the list of loaded keyman keyboards.
'
'    To disable a keyboard, call:
'          SelectKeymanKeyboard "<name>", False
'
'    To start Keyman from VBA, call:
'          StartKeyman
'
'    To retrieve a list of all loaded keyboards, use:
'          i = 0
'          Do
'              i = i+1
'              name(i) = KeyboardName(i)
'          Loop While name(i) <> ""
'
'    To determine the currently active keyboard name, use (for example):
'          MsgBox ActiveKeymanKeyboard
'    The string "(None)" will be returned if no keyboard is active, and an empty string
'    will be returned if Keyman is not loaded.
'
'    Routines to test if Keyman is running and to check the version are also provided,
'    but should not normally be required.
'
'  Warning:
'    This code has been optimized for making keyboard changes quickly, but may fail
'    if keyboards are loaded or unloaded after it has been initialized. If this is likely
'    to be a problem, include the explicit call:
'          InitializeKeyman True
'    before any other function call. Including this argument will ensure that keyboard
'    switching is safe, but will also slow it down.
'
'  Note: Updated for Keyman Desktop 7.0 on 20 March 2007
'
'    Copyright 2007 Tavultesoft Pty Ltd http://www.tavultesoft.com
'
'###########################################################################################

Option Base 1
Option Explicit

' Windows API function prototypes

Private Declare Function Sleep Lib "Kernel32" (ByVal mSec As Long) As Long
Private Declare Function lstrcpyW Lib "Kernel32" (ByVal lpString1 As Any, ByVal lpString2 As Any) As Long
Private Declare Function lstrlenW Lib "Kernel32" (ByVal lpString1 As Any) As Long

Private Declare Function FindWindow Lib "User32" Alias "FindWindowA" (ByVal clName As Any, ByVal wName As Any) As Long
Private Declare Function RegisterWindowMessage Lib "User32" Alias "RegisterWindowMessageA" (ByVal MessageID As String) As Long
Private Declare Function GetFocus Lib "User32" () As Long
Private Declare Function PostMessage Lib "User32" Alias "PostMessageA" (ByVal hWnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long

' Windows registry functions, constants and structures

Public Type SECURITY_ATTRIBUTES
    nLength As Long
    lpSecurityDescriptor As Long
    bInheritHandle As Boolean
End Type

Private Declare Function RegCloseKey Lib "advapi32.dll" (ByVal hKey As Long) As Long
Private Declare Function RegOpenKeyEx Lib "advapi32.dll" Alias "RegOpenKeyExA" (ByVal hKey As Long, ByVal lpSubKey As String, ByVal ulOptions As Long, ByVal samDesired As Long, phkResult As Long) As Long
Private Declare Function RegQueryValueExString Lib "advapi32.dll" Alias "RegQueryValueExA" (ByVal hKey As Long, ByVal lpValueName As String, ByVal lpReserved As Long, lpType As Long, ByVal lpData As String, lpcbData As Long) As Long

Const HKROOT = &H80000000
Const HKCU = &H80000001
Const HKLM = &H80000002
Const REG_SZ = 1
Const KEY_READ As Long = &H20019

' Keyman API prototypes, constants and structures

Private Declare Function Keyman_BuildKeyboardList Lib "keyman32" (ByRef pKbdInfo As Any, ByRef pCount As Long) As Long
Private Declare Function GetSystemStore Lib "keyman32" (ByVal pKeybd As Long, ByVal Index As Long) As Long
Private Declare Function GetActiveKeymanID Lib "keyman32" () As Long

Const RWM_KMSELECTLANG = "WM_KMSELECTLANG"  'Keyman registered message
Const KEYMANID_NONKEYMAN = &HFFFFFFFF       'Keyboard default (off) ID
Const TSS_NAME = 7                          'GetSystemStore index value

Const skKeymanID = 1                        'Keyboard selection mode constant

Type KEYBOARDINFO       'Structure is correct for Version 5, and for Version 6.0.154 (or beta) and later
    KeymanID As Long
    HotKey As Long
    KeyboardID As Long
    KbdName(1 To 256) As Byte
    pKeyboard As Long
    nIMDLLs As Long
    IMDLLs As Long
End Type

' VBA Interface constants, variables, structures and global functions

Const sNullString = 0&                      'VBA constant for passing a null pointer

Type KEYMANKEYBOARD                         'VBA type for holding keyboard names and IDs
    KbdName As String
    ID As Long
End Type

' (Static) variables set during initialization

Dim wm_kmselectlang As Long, nLayouts As Long
Dim Keyboards() As KEYMANKEYBOARD

Public Function KeymanOn() As Boolean
'
' Check if Keyman 6 (or 5) is executing
'
    KeymanOn = (FindWindow(sNullString, "Keyman50") <> 0) Or (FindWindow(sNullString, "frmKeyman7Main") <> 0)
End Function

Public Function GetKeymanVersion(Optional Path)
'
' Check the Keyman version in the registry and optionally return the path of keyman.exe
'
    Dim RootPath As String, EXEPath As String
    RootPath = GetRegStringValue(HKLM, "SOFTWARE\Tavultesoft\keyman\7.0", "root path")
    If RootPath <> "" Then
        EXEPath = RootPath & "kmshell.exe"
        If Dir(EXEPath) <> "" Then
            GetKeymanVersion = 70
            If Not IsMissing(Path) Then Path = EXEPath
            Exit Function
        End If
    End If

    RootPath = GetRegStringValue(HKLM, "SOFTWARE\Tavultesoft\keyman\6.0", "root path")
    If RootPath <> "" Then
        EXEPath = RootPath & "keyman.exe"
        If Dir(EXEPath) <> "" Then
            GetKeymanVersion = 60
            If Not IsMissing(Path) Then Path = EXEPath
            Exit Function
        End If
    End If
    
    RootPath = GetRegStringValue(HKLM, "SOFTWARE\Tavultesoft\keyman\5.0", "root path")
    If RootPath <> "" Then
        EXEPath = RootPath & "keyman.exe"
        If Dir(EXEPath) <> "" Then
            GetKeymanVersion = 50
            If Not IsMissing(Path) Then Path = EXEPath
            Exit Function
        End If
    End If

    GetKeymanVersion = 0        ' missing or invalid registry key
    If Not IsMissing(Path) Then Path = ""
    
End Function

Public Function StartKeyman() As Boolean
'
' Start Keyman if it is not yet executing
'
    Dim tWait As Long, EXEPath As String
    
    StartKeyman = False

    If KeymanOn() Then StartKeyman = True: Exit Function

    Dim v As Long
    v = GetKeymanVersion(EXEPath)
    If v < 60 Or v > 70 Then Exit Function
    
    If Shell(EXEPath) = 0 Then Exit Function
        
    ' Wait until Keyman is executing, then check again
    Sleep 500
    For tWait = 500 To 5000 Step 500
        If KeymanOn() Then
            StartKeyman = True
            Sleep 2500
            Exit Function
        End If
    Next
End Function

Public Function InitializeKeyman(Optional ForceReload As Boolean = False) As Boolean
'
' Load the Keyman library, initialize address pointers and messages, and load keyboards into memory
'
    Dim KeybdInfo() As KEYBOARDINFO, pKbdAdr As Long, s As String, tAdr As Long, nChar As Long, _
        j As Integer, n As Integer, t As String
    Static Initialized As Boolean
    
    If ForceReload Then Initialized = False
    
    ' Keyman must be running - but don't start it here to avoid possible conflicts
    InitializeKeyman = False
    If Not KeymanOn() Then Exit Function    'KeymanOn() can only return true for Keyman 5 or later
    
    If Initialized Then
        InitializeKeyman = nLayouts
        Exit Function
    End If
    
    ' Register Keyman message
    wm_kmselectlang = RegisterWindowMessage(RWM_KMSELECTLANG)
    
    If Keyman_BuildKeyboardList(ByVal 0&, nLayouts) = 0 Then Exit Function
    
    If nLayouts = 0 Then Exit Function

    ReDim Keyboards(nLayouts)
    ReDim KeybdInfo(nLayouts)
    If Keyman_BuildKeyboardList(KeybdInfo(1), nLayouts) = 0 Then Exit Function

    ' Get actual keyboard name where possible
    For n = 1 To nLayouts
        s = ""
        
        'and check for a valid structure address
        pKbdAdr = KeybdInfo(n).pKeyboard
        If pKbdAdr <> 0 Then
            
            'then get address of the (Unicode) string with the keyboard name
            tAdr = GetSystemStore(pKbdAdr, TSS_NAME)
            nChar = lstrlenW(tAdr)
            
            'copy the string to a temporary buffer
            s = Space(2 * nChar + 32)
            tAdr = lstrcpyW(s, tAdr)
            
            'and convert it to an ASCII string
            t = ""
            For j = 1 To nChar
                t = t & Mid(s, 2 * j - 1, 1)
            Next
        Else
            'if no valid keyboard address, just pack the file name
            t = ""
            For j = 1 To 256
                If KeybdInfo(n).KbdName(j) = 0 Then Exit For
                t = t & Chr(KeybdInfo(n).KbdName(j))
            Next
        End If
        Keyboards(n).ID = KeybdInfo(n).KeymanID
        Keyboards(n).KbdName = t
    Next
        
    InitializeKeyman = True
    Initialized = True
    
End Function

Public Sub SelectKeymanKeyboard(Optional pKbdName, Optional Enable As Boolean = True)
'
' Select keyboard by name for enabling (or disabling)
'
    Dim hWnd As Long, KeyboardNo As Long
        
    If InitializeKeyman() Then
        KeyboardNo = KEYMANID_NONKEYMAN
        If IsMissing(pKbdName) Or Not Enable Then
            PostMessage GetFocus(), wm_kmselectlang, skKeymanID, KEYMANID_NONKEYMAN
        Else
            PostMessage GetFocus(), wm_kmselectlang, skKeymanID, KeymanID(pKbdName)
        End If
    End If
End Sub

Public Function KeyboardName(KeybdNo) As String
'
' Look up a keyboard name (according to array order, not ID)
'
    Dim n As Long
    
    KeyboardName = ""
    If InitializeKeyman() Then
        If KeybdNo > 0 And KeybdNo <= nLayouts Then
            KeyboardName = Keyboards(KeybdNo).KbdName
        End If
    End If
End Function

Private Function KeymanID(KbdName) As Long
'
' Look up a keyboard number from the name
'
    Dim n As Long
    
    If InitializeKeyman() Then
        For n = 1 To nLayouts
            If Keyboards(n).KbdName = KbdName Then
                KeymanID = Keyboards(n).ID
                Exit Function
            End If
        Next
    End If
    KeymanID = -1
End Function

Public Function ActiveKeymanKeyboard() As String
'
' Look up a keyboard number from the name
'
    Dim n As Long, ActiveKeyboardID As Long
    
    If InitializeKeyman() Then
        ActiveKeyboardID = GetActiveKeymanID()
        If ActiveKeyboardID <> -1 Then
            For n = 1 To nLayouts
                If ActiveKeyboardID = Keyboards(n).ID Then
                    ActiveKeymanKeyboard = Keyboards(n).KbdName
                    Exit Function
                End If
            Next
        Else
            ActiveKeymanKeyboard = "(None)"
        End If
    Else
        ActiveKeymanKeyboard = ""
    End If
End Function

Private Function GetRegStringValue(HKeyBase, KeyString As String, ValueName As String, Optional Default) As String
    Dim OutStr As String, OutType As Long, OutLen As Long, hRegKey As Long

    If IsMissing(Default) Then GetRegStringValue = "" Else GetRegStringValue = CStr(Default)

    If RegOpenKeyEx(HKeyBase, KeyString, 0&, KEY_READ, hRegKey) <> 0 Then Exit Function

    OutLen = 512
    OutStr = String(512, 0)

    If RegQueryValueExString(hRegKey, ValueName, 0&, OutType, OutStr, OutLen) <> 0 Then Exit Function

    RegCloseKey hRegKey

    If OutType <> REG_SZ Then Exit Function

    GetRegStringValue = Left(OutStr, InStr(OutStr, Chr(0)) - 1)
End Function
